# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import collections
import itertools
import logging
import time
from urllib.parse import urlparse, urlunparse

import requests
from django.core.handlers.base import BaseHandler
from django.test.client import RequestFactory
from requests.adapters import BaseAdapter
from requests.cookies import extract_cookies_to_jar
from requests.exceptions import ConnectionError, Timeout
from requests.models import RequestEncodingMixin, Response
from requests.structures import CaseInsensitiveDict
from requests.utils import get_encoding_from_headers

from rest_models.backend.exceptions import FakeDatabaseDbAPI2
from rest_models.backend.utils import message_from_response

logger = logging.getLogger("django.db.backends")


def build_url(url, params):
    """
    only for display purpose, parse the params and url to build the final api
    :param str url: the url path to use
    :param dict params: the dict with the GET parameters, as accepted by requests
    :return: the url with the get part
    :rtype: str
    """
    if params is None:
        result = url
    else:
        result = "%s?%s" % (url, RequestEncodingMixin._encode_params(params))
    return result


class LocalApiAdapter(BaseAdapter):

    SPECIAL_URL = "http://localapi"

    def __init__(self):
        self.request_factory = RequestFactory()
        super(LocalApiAdapter, self).__init__()

    def prepared_request_to_wsgi_request(self, prepared_request):
        """
        transform a PreparedRequest into a WsgiRequest for django to use
        :param requests.models.PreparedRequest prepared_request:
        :return: the request ready to use for django
        :rtype: django.core.handlers.wsgi.WSGIRequest
        """
        wsgi_request = self.request_factory.generic(
            method=prepared_request.method,
            path=prepared_request.url,
            data=prepared_request.body,
            content_type=prepared_request.headers.get('Content-Type', 'application/x-www-form-urlencoded')
        )
        for name, val in prepared_request.headers.items():
            wsgi_request.META[str('HTTP_') + str(name.upper())] = val
        return wsgi_request

    def http_response_to_response(self, http_response, prepared_request):
        """
        transform a WSGIResponse into a requests's Response model
        :param django.http.response.HttpResponse|django.http.response.StreamingHttpResponse http_response:
            the http response send by django view
        :return: the requests's Response model corresponding to the http_response
        :rtype: Response
        """
        response = Response()

        # Fallback to None if there's no status_code, for whatever reason.
        response.status_code = getattr(http_response, 'status_code', None)

        # Make headers case-insensitive.
        response.headers = CaseInsensitiveDict(getattr(http_response, 'headers', {}))

        # Set encoding.
        response.encoding = get_encoding_from_headers(response.headers)
        response.raw = http_response
        response.reason = response.raw.reason_phrase
        try:
            response._content = http_response.content
        except AttributeError:
            response._content = http_response.getvalue()
            response.raw.read = lambda *args: http_response.getvalue()
        req = prepared_request

        if isinstance(req.url, bytes):  # pragma: no cover
            response.url = req.url.decode('utf-8')
        else:
            response.url = req.url

        # Add new cookies from the server.
        extract_cookies_to_jar(response.cookies, req, response)

        # Give the Response some context.
        response.request = req
        response.connection = self

        return response

    def send(self, request, stream=False, timeout=None, verify=True,
             cert=None, proxies=None):
        handler = BaseHandler()
        handler.load_middleware()
        wsgi_request = self.prepared_request_to_wsgi_request(request)
        http_response = handler.get_response(wsgi_request)
        requests_response = self.http_response_to_response(http_response, request)
        return requests_response

    def close(self):  # pragma: no cover
        pass


class ApiVerbShortcutMixin(object):

    def get(self, url, params=None, json=None, **kwargs):
        """
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes json: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :return:
        :rtype: requests.Response
        """
        return self.request("get", url, params=params, json=json, **kwargs)

    def head(self, url, params=None, json=None, **kwargs):
        """
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes json: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :return:
        :rtype: requests.Response
        """
        return self.request("head", url, params=params, json=json, **kwargs)

    def post(self, url, params=None, json=None, **kwargs):
        """
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes json: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :return:
        :rtype: requests.Response
        """
        return self.request("post", url, params=params, json=json, **kwargs)

    def put(self, url, params=None, json=None, **kwargs):
        """
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes json: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :return:
        :rtype: requests.Response
        """
        return self.request("put", url, params=params, json=json, **kwargs)

    def patch(self, url, params=None, json=None, **kwargs):
        """
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes json: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :return:
        :rtype: requests.Response
        """
        return self.request("patch", url, params=params, json=json, **kwargs)

    def options(self, url, params=None, json=None, **kwargs):
        """
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes json: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :return:
        :rtype: requests.Response
        """
        return self.request("options", url, params=params, json=json, **kwargs)

    def delete(self, url, params=None, json=None, **kwargs):
        """
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes json: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :return:
        :rtype: requests.Response
        """
        return self.request("delete", url, params=params, json=json, **kwargs)


class DebugApiConnectionWrapper(ApiVerbShortcutMixin):
    def __init__(self, connection, db):
        self.connection = connection
        self.db = db

    def __getattr__(self, attr):  # pragma: no cover
        cursor_attr = getattr(self.connection, attr)
        return cursor_attr

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.connection.close()

    # django-debug-toolbar>=4 compatibility
    @property
    def cursor(self):
        return self.connection

    def execute(self, *args, **kwargs):
        return self.connection.execute(*args, **kwargs)

    def request(self, method, url, **kwargs):

        start = time.time()
        response = None
        try:
            response = self.connection.request(method, url, **kwargs)
        finally:
            stop = time.time()
            duration = stop - start
            sql = build_url(url, kwargs['params'])
            elapsed_sec = response.elapsed.total_seconds() if response else 0.
            self.db.queries_log.append({
                'sql': "%s %s ||| %s" % (method, sql, kwargs),
                'time': "%.3f " % elapsed_sec
            })
            logger.debug('(%.3f) (backend:%.3f) %s %s; args=%s' % (duration, elapsed_sec,
                                                                   method, sql, kwargs),
                         extra={'duration': duration, 'backend': elapsed_sec,
                                'sql': sql, 'params': kwargs, 'method': method}
                         )
        return response


def get_basic_session():
    session = requests.Session()
    session.mount(LocalApiAdapter.SPECIAL_URL, LocalApiAdapter())
    session.mount('http://testserver', LocalApiAdapter())  # for special host used by django for unittests
    return session


class ApiConnexion(ApiVerbShortcutMixin):
    """
    wrapper for request.Session that in fact implement useless methods like rollback which
    is not possible with a rest API
    """
    def __init__(self, url, auth=None, retry=3, timeout=3, backend=None, middlewares=(), ssl_verify=None):
        """
        create a persistent connection to the api
        :param str url: the base url for the api (host + port + start path)
        :param ApiAuthBase auth: the auth provider
        :param int retry: number of retry
        :param int|(int, int) timeout: the timeout to pass to the api
        :param DatabaseWrapper backend: the backend
        :param list[Middleware]|tuple[Middleware] middlewares:the list of middleware to execute for each query.
        """
        if not url.endswith('/'):
            # fix the miss configured url in the api (must end with a /)
            url = url + '/'
        self.session = requests.Session()
        self.session.mount(LocalApiAdapter.SPECIAL_URL, LocalApiAdapter())
        self.session.mount('http://testserver', LocalApiAdapter())  # for special host used by django for unittests
        self.session.auth = self.auth = auth
        self.url = url
        self.retry = retry
        self.timeout = timeout
        self.backend = backend
        self._middlewares_scheduler = collections.defaultdict(list)
        self._requestid = 0
        if ssl_verify is not None:
            self.session.verify = ssl_verify
        for middleware in middlewares:
            self.push_middleware(middleware, 8)

    @property
    def middlewares(self):
        """
        the list of middleware to iterate over. ordered by priority from 1 to 10
        :return:
        """
        return list(itertools.chain(*(v for k, v in sorted(self._middlewares_scheduler.items()))))

    def push_middleware(self, middleware, priority=5):
        """
        push a middleware into the list, which will be called at each request and response processing
        :param ApiMiddleware middleware: the middleware to apppend
        :return:
        """
        self._middlewares_scheduler[priority].insert(0, middleware)

    def pop_middleware(self, middleware):
        for middlewares in self._middlewares_scheduler.values():
            try:
                middlewares.remove(middleware)
            except ValueError:
                pass

    def close(self):
        self.session.close()

    def __exit__(self, *args):
        self.close()

    def __enter__(self):
        return self

    def inc_request_id(self):
        """
        increment the request id and then return it
        :return:
        """
        self._requestid += 1
        return self._requestid

    def execute(self, sql, params=None):
        """

        :param sql: the usless sql
        :param params: the userless params
        :return:
        """
        params = params or {}
        return self._make_request(params)

    def _make_request(self, params):
        """
        finaly make the request. will pass all middleware on send and recieve
        :param params:
        :return:
        """
        requestid = self.inc_request_id()
        middlewares = self.middlewares
        i = 0
        for i, middleware in enumerate(middlewares):
            response = middleware.process_request(params, requestid, self)
            if response is not None:
                break
        else:
            # if the middleware did not override the real response, we make the query
            response = self.session.request(**params)
        for middleware in middlewares[i::-1]:  # iterate over all previously executed middlewares
            response = middleware.process_response(params, response, requestid) or response
        return response

    def rollback(self):  # pragma: no cover
        pass

    def commit(self):  # pragma: no cover
        pass

    def get_timeout(self):
        return self.timeout

    def request(self, method, url, **kwargs):
        """
        wrapper for requests.Session.get

        :param unicode method: the method to use
        :param str url: the path relative to the current connexion
        :param dict[str, any]|byes params: (optional) Dictionary or bytes to append as GET parameters
        :param dict[str, any]|bytes data: (optional) Dictionary, bytes, or file-like object to send
            in the body
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param files: (optional) Dictionary of ``'filename': file-like-objects``
            for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param float|tuple[float, float] timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :rtype: requests.Response

        """
        kwargs.setdefault("allow_redirects", False)
        kwargs.setdefault("timeout", self.get_timeout())
        kwargs.setdefault('stream', False)
        real_url = self.get_final_url(url)

        error = 0
        last_exception = None

        while error <= self.retry:
            try:
                # to stay compatible with django_debug_toolbar, we must
                # call execute on the cursor return by the backend, since this one is replaced
                # by django-debug-toolbar to state the query.
                if self.backend:
                    execute = self.backend.cursor().execute
                else:
                    execute = self.execute
                response = execute("%s %s" % (method.upper(), real_url), dict(method=method, url=real_url, **kwargs))

            except Timeout as e:
                error += 1
                last_exception = e
            except ConnectionError as e:
                error += 1
                last_exception = e
            else:

                if self.auth and hasattr(self.auth, 'raise_on_response_forbidden'):
                    self.auth.raise_on_response_forbidden(response)
                else:
                    if response.status_code in (403, 401):
                        raise FakeDatabaseDbAPI2.ProgrammingError(
                            "Access to database is Forbidden for user %s.\n%s" %
                            (self.auth[0] if isinstance(self.auth, tuple) else self.auth,
                             message_from_response(response))
                        )
                return response
        raise FakeDatabaseDbAPI2.OperationalError(
            "cound not connect to server: %s\nIs the API running on %s ? tried %d times" %
            (last_exception, self.url, error))

    def get_final_url(self, url):
        if url.startswith("/"):
            parsed_url = urlparse(url)
            api_url = urlparse(self.url)
            return urlunparse(api_url[:2] + parsed_url[2:])

        return self.url + url



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import json
from collections import defaultdict
from pathlib import Path


def dict_contains(subdict, maindict):
    """
    return True if the subdict is present with the sames values in dict.
    can be recursive. if maindict contains some key not in subdict it's ok.
    but if subdict has a key not in maindict or the value are not the same, it's a failure.

    >>> dict_contains(dict(a=1, b=2), dict(a=1, c=2, b=2))
    True

    >>> dict_contains(dict(a=dict(aa=1)), dict(a=dict(aa=1, bb=2), b=2))
    True

    >>> dict_contains(dict(a=dict(aa=1, bb=2)), dict(a=dict(aa=2)))
    False

    >>> dict_contains(dict(a=dict(aa=1)), dict(a=[]))
    False

    >>> dict_contains(dict(a=1), dict())
    False

    >>> dict_contains(dict(a=[1, 2, 3]), dict(a=[1, 2, 3]))
    True

    >>> dict_contains(dict(a=[1, 3, 2]), dict(a=[1, 2, 3]))
    False

    >>> dict_contains(dict(a=[1, 3]), dict(a=[1, 2, 3]))
    False

    >>> dict_contains(dict(a=[1, 3, 2]), dict(a={1, 2, 3}))
    True

    :param subdict: the smaller dict that should be present in the big one
    :param maindict: the dict
    :return: True if subdict is included in maindict
    :rtype: bool
    """
    try:
        for k, v in subdict.items():
            mainv = maindict[k]
            if isinstance(mainv, dict) and isinstance(v, dict):
                if not dict_contains(v, mainv):
                    return False
            elif isinstance(mainv, (set, frozenset)):
                return set(v) == mainv
            elif mainv != v:
                return False
    except KeyError:
        return False
    return True


class JsonFixtures(object):
    """
    a class that will take raw data or filenames to load api fixtures.
    it work with many usefull synthax.

    JsonFixtures(
        nested_json_fixtures,
        'path/to/fixtures.json',  # this fixture will contains the urls

        'model/url/'=[response...],
        'model_url': 'path/to/other/fixture.json  # this fixtures is a list of response
    )

    some variables can be injected in the urls::

    >>> f = JsonFixtures({'/model/%(pk)s/': ['hey']})
    >>> f['/model/134/'] # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    KeyError: ...'/model/134/'
    >>> f.set_variable({'pk': 134}) and None
    >>> f['/model/134/'] == ['hey']
    True

    >>> f = JsonFixtures({'/model': ['a'], 'b': 2}, c=3)
    >>> f.update({'d': 4, '/model': ['b']}, c=4)
    >>> f['/model'] == ['b', 'a']
    True
    >>> f['b']
    [2]
    >>> f['c']
    [4, 3]
    >>> f['d']
    [4]
    """

    def __init__(self, *args, **kwargs):
        self.files = []
        self.url_for_data = defaultdict(list)
        self.variable = {}
        self.update(*args, **kwargs)

    def update(self, *args, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, (list, tuple)):
                self.url_for_data[k][0:0] = list(v)
            else:
                self.url_for_data[k].insert(0, v)
        for file in args:
            if isinstance(file, (str, bytes, Path)):
                self.files.append(file)
            elif isinstance(file, dict):
                self.update(**file)
            elif isinstance(file, JsonFixtures):
                self.update(**file._load())
            else:
                raise ValueError("%s is not supported as *args" % (type(file)))

    def set_variable(self, variable):
        """
        set the dict that will be used as variable for the url resolving
        :param variable:
        :return:
        """
        self.variable = variable
        return self

    def __load_json(self, path):
        """
        helper that load the json in path
        :param str|Path path: the path to load
        :return: the data loaded
        """
        try:
            with Path(path).open('r') as f:
                return json.load(f)
        except ValueError as ve:
            raise ValueError("error while loading the fixture %s" % path) from ve

    def _load(self):
        """
        load the data from files and url_for_data and return the dict as expected by JsonFixtures
        :rtype: dict
        """
        res = {}
        for file in self.files:
            loaded = self.__load_json(file)
            for k, v in loaded.items():
                if isinstance(v, (list, tuple)):
                    res.setdefault(k, []).extend(v)
                else:
                    res.setdefault(k, []).append(v)

        for url, datas in self.url_for_data.items():
            if not isinstance(datas, (list, tuple)):
                datas = [datas]
            for data in datas:
                res.setdefault(url, [])
                if isinstance(data, Path):
                    data = self.__load_json(data)
                if isinstance(data, (list, tuple)):
                    res[url].extend(data)
                else:
                    res[url].append(data)
        return res

    def __getitem__(self, item):
        if not hasattr(self, "_loaded"):
            self._loaded = self._load()
        try:
            return self._loaded[item]
        except KeyError as ke:
            # try with varables
            for url, data in self._loaded.items():
                try:
                    if url % self.variable == item:
                        return data
                except KeyError:
                    # the resolving of url can use a %(name)s that is not in the dict
                    pass
            raise ke

    def items(self):
        if not hasattr(self, "_loaded"):
            self._loaded = self._load()
        for url, data in self._loaded.items():
            try:
                yield url % self.variable, data
            except KeyError:
                # the resolving of url can use a %(name)s that is not in the dict
                pass


def pgcd(a, b):
    """
    return the best page size for a given limited query

    :param a: the start offset
    :param b: the end offset
    :return:

    >>> pgcd(30, 40)
    10

    >>> pgcd(7, 13)
    1

    """
    while a % b != 0:
        a, b = b, a % b
    return b



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import logging
from importlib import import_module

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import connections

from rest_models.backend.base import DatabaseWrapper

logger = logging.getLogger(__name__)


def get_default_api_database(databases):
    """
    parse all database settings and return the database matching current backend.
    raise ImproperlyConfigured if there is none or many backend.
    :return:
    """
    _api_database_name = None
    for db_name, db_settings in databases.items():
        # prevent the special alias "TEST_" database to be used.
        # in test environments, the value of the TEST_* is copied into the
        # original one.
        if RestModelRouter.is_restmodel_database(db_settings) and not db_name.startswith('TEST_'):

            if _api_database_name is None:

                _api_database_name = db_name
            else:
                raise ImproperlyConfigured("too many Api Database found. you must specify "
                                           "the database to use in each model.APIMeta.db_name")
    if _api_database_name is None:
        raise ImproperlyConfigured("no Api Database found in settings.DATABASE. you can't use a model with "
                                   "APIMeta in INSTALLED_APPS application")
    return _api_database_name


class RestModelRouter(object):

    def __init__(self):
        self._api_database_name = None
        self.cache = {}
        self.databases = settings.DATABASES

    @staticmethod
    def is_restmodel_database(db_settings):
        db_module = import_module(db_settings['ENGINE'] + '.base')
        return issubclass(db_module.DatabaseWrapper, DatabaseWrapper)

    @property
    def api_database_name(self):
        if self._api_database_name is None:
            self._api_database_name = get_default_api_database(self.databases)
        return self._api_database_name

    @staticmethod
    def is_api_model(model):
        return hasattr(model, 'APIMeta')

    def get_api_database(self, model):
        try:
            return self.cache[model]
        except KeyError:
            if self.is_api_model(model):
                if hasattr(model.APIMeta, 'db_name'):
                    result = model.APIMeta.db_name
                else:
                    result = self.api_database_name
            else:
                result = None
            self.cache[model] = result
            return result

    def get_api_connexion(self, model):
        """
        return the connection to query the api behind this model
        :param model:
        :return: the api connection
        :rtype: rest_models.backend.base.DatabaseWrapper
        """
        return connections[self.get_api_database(model)]

    def db_for_read(self, model, **hints):
        return self.get_api_database(model)

    def db_for_write(self, model, **hints):
        return self.get_api_database(model)

    def allow_relation(self, obj1, obj2, **hints):
        if self.is_api_model(obj1) ^ self.is_api_model(obj2):
            return False
        return None

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if self.is_restmodel_database(self.databases[db]):
            return False  # no migration for our database

        if model_name is None:
            return None  # we are model specific
        # all our models must not being created in other databases
        try:
            model = apps.get_model(app_label, model_name)
        except LookupError:
            return None
        if self.is_api_model(model):
            return False
        return None



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import logging

from django.db.backends.base.operations import BaseDatabaseOperations

logger = logging.getLogger(__name__)


class DatabaseOperations(BaseDatabaseOperations):
    compiler_module = 'rest_models.backend.compiler'
    # postgis compatibility
    select = '%s'

    def sql_flush(self, *args, **kwargs):  # pragma: no cover
        return ""

    def quote_name(self, name):
        return "quoted(!%s!)" % name

    def adapt_json_value(self, value, encoder):
        return value



from __future__ import unicode_literals

from django.db.backends.base.features import BaseDatabaseFeatures


class DatabaseFeatures(BaseDatabaseFeatures):
    # SQLite cannot handle us only partially reading from a cursor's result set
    # and then writing the same rows to the database in another cursor. This
    # setting ensures we always read result sets fully into memory all in one
    # go.
    can_use_chunked_reads = True
    test_db_allows_multiple_connections = True
    supports_unspecified_pk = True
    supports_timezones = False
    supports_1000_query_parameters = False
    supports_mixed_date_datetime_comparisons = False
    has_bulk_insert = True
    can_return_id_from_insert = True
    can_combine_inserts_with_and_without_auto_increment_pk = False
    supports_foreign_keys = True
    supports_column_check_constraints = False
    autocommits_when_autocommit_is_off = False
    can_introspect_decimal_field = False
    can_introspect_positive_integer_field = True
    can_introspect_small_integer_field = True
    supports_transactions = False
    atomic_transactions = False
    can_rollback_ddl = False
    supports_paramstyle_pyformat = False
    supports_sequence_reset = False
    has_select_for_update_skip_locked = False

    uses_savepoints = False

    can_release_savepoints = False
    can_share_in_memory_db = False
    supports_stddev = False

    has_zoneinfo_database = False



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import logging

logger = logging.getLogger(__name__)


def message_from_response(response):
    return "[%d]%s" % (
        response.status_code,
        response.text if '<!DOCTYPE html>' not in response.text[:30] else response.reason
    )


try:
    from django.db.models import JSONField as JSONFieldLegacy
except ImportError:
    pass

try:
    class JSONField(JSONFieldLegacy):
        def get_prep_value(self, value):
            return value

        def from_db_value(self, value, expression, connection):
            try:
                return super().from_db_value(value, expression, connection)
            except TypeError:
                return value

except NameError:
    def JSONField(*args, **kwargs):
        return None



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import datetime
import logging
import os
import threading
from urllib.parse import unquote

import unidecode
from django.core.files.base import ContentFile
from django.core.files.storage import Storage
from django.utils.deconstruct import deconstructible

from rest_models.backend.connexion import get_basic_session

logger = logging.getLogger(__name__)


class ExpirableDict(dict):

    def __init__(self, maxage=datetime.timedelta(hours=1)):
        self.maxage = maxage
        self.rlock = threading.RLock()
        super(ExpirableDict, self).__init__()

    def _clean_cache(self):
        latest = datetime.datetime.now() - self.maxage

        with self.rlock:
            to_clear = [
                key
                for key, (d, val) in self.items()
                if d < latest
            ]
            for k in to_clear:
                del self[k]

    def __delitem__(self, key):
        with self.rlock:
            super(ExpirableDict, self).__delitem__(key)

    def __getitem__(self, key):
        with self.rlock:
            return super(ExpirableDict, self).__getitem__(key)[1]

    def get(self, key, default=None):
        with self.rlock:
            if key in self:
                return super(ExpirableDict, self).get(key)[1]
            else:
                return default

    def pop(self, key, default=None):
        with self.rlock:
            return super(ExpirableDict, self).pop(key, default)

    def __setitem__(self, key, value):
        with self.rlock:
            super(ExpirableDict, self).__setitem__(key, (datetime.datetime.now(), value))
        self._clean_cache()


class RestFileField(str):
    content = None

    @property
    def name(self):
        return self.content.name if self.content else ''

    def __init__(self, content):
        self.content = content

    def __str__(self):
        return self.content and self.content.name


@deconstructible
class RestApiStorage(Storage):
    def __init__(self):
        self.result_file_pool = ExpirableDict()

    def prepare_result_from_api(self, result, cursor):
        # result is the full url from the api, we will return only the name,
        # and store the full url for later
        if result is None:
            return None
        name = unquote(os.path.basename(result))
        self.result_file_pool[name] = result, cursor
        return name

    def _open(self, name, mode='rb'):
        if 'r' not in mode:
            NotImplementedError("This backend doesn't support writing on file.")
        cursor = self.get_cursor(name)  # fetch a valid cursor which just got the name
        url = self.url(name)
        # try with anonymous data
        response = get_basic_session().get(url)
        # if we fail, we try with authenticated session
        if response.status_code in (403, 401):
            response = cursor.session.get(url)
        response.raise_for_status()
        return ContentFile(response.content, name)

    @staticmethod
    def _save(name, content):
        # self.uploaded_file_pool[id] = content
        content.name = content.original_name = name
        return RestFileField(content)

    def size(self, name):
        cursor = self.get_cursor(name)  # fetch a valid cursor which just got the name
        url = self.url(name)
        # try with anonymous data
        response = get_basic_session().get(url, stream=True)
        # if we fail, we try with authenticated session
        if response.status_code in (403, 401):
            response = cursor.session.get(url, stream=True)

        return response.headers['Content-length']

    def get_cursor(self, name):
        return self.result_file_pool[name][1]  # pool contains url and cursor to get it

    def url(self, name):
        """return a usable url from this name"""
        # since the api return us a full url, and it's considered just the name by django
        # we don't do anything

        return self.result_file_pool.get(name, (name, None))[0]  # pool contains url and cursor to get it

    @staticmethod
    def get_available_name(name, max_length=None):
        return unidecode.unidecode(name)

    @staticmethod
    def get_valid_name(name):
        return unidecode.unidecode(name)
